#|<--                          80 columns                                   ->|#

(let ((time-stamp "Time-stamp: <03/05/19 17:10:34 kazu>"))
;;  (display (string-append "parser.scm: " (substring time-stamp 13 29)))
  (display (string-append "parser.scm: " (substring time-stamp 13 30)))
  (newline))

;;;
;;; conventions:
;;;
;;;  Anything that can be represented as a symbol, number or list is
;;;  represented that way.  BUT, it is never called symbol or list or
;;;  syms or anything like that, it is called by the name of the thing
;;;  it represents.  Examples include:
;;;
;;;      the 'name' of a param
;;;      the 'params' of a procedure
;;;      the 'exps' of a body
;;;

(define parse-program
  (lambda (sexp) 
    (make-program (parse-decls sexp))))

(define parse-decls
  (lambda (sexps)
    (let ((parse-one parse-decl))
      (remove* #f (map parse-one sexps))))) 

(define parse-member-decls
  (lambda (enclosing-cname sexps)
    (let ((parse-one (lambda (sexp) (parse-member-decl enclosing-cname sexp))))
      (remove* #f (map parse-one sexps)))))

;; parse "declaration" = "class declaration"
(define parse-decl
  (lambda (sexp)
    (if (not (and (list? sexp)
		  (eqv? (car sexp) 'class)))
	(error 'parse-decl "Not a decl -- ~s" sexp)
	;; case : sexp is a list and its head is 'class'
	(make-class-decl (cadr sexp) 
			 (caddr sexp)
			 (parse-member-decls (cadr sexp) (cdddr sexp))))))

;; parse "member declaration" in "class declaration"
;;   member is "field" or "method"
(define parse-member-decl
  (lambda (enclosing-cname sexp)
    (or (and (list? sexp)
	     (case (car sexp) 
	       ;;
	       ;; (field int x)
	       ;;
	       ((field)
		(make-field-decl enclosing-cname (cadr sexp) (caddr sexp)))
	       
	       ;;
	       ;; (method void setX ((int nx)) (fset! this x nx))
	       ;;
	       ((method)
		(if (procedure? (cadddr sexp))
		    (make-method-decl enclosing-cname
				      (cadr sexp) (caddr sexp) '()
				      (cadddr sexp)) ;ESCAPE to Scheme
		    (make-method-decl enclosing-cname
				      (cadr sexp)
				      (caddr sexp) ;for the sake of open classes this
					           ;is allowed to be a list
				      (parse-params (cadddr sexp))
				      (parse-body (cddddr sexp)))))))
	(error 'parse-member-decl "Not a member decl -- ~s" sexp))))


;; parse "parameter" : parameter has "type" and "pname"
(define parse-param  
  (lambda (sexp)  
    (make-param (car sexp) (cadr sexp))))

(define parse-params 
  (lambda (sexps) 
    (map parse-param sexps)))

;; parse "method body" in "method declaration"
(define parse-body   
  (lambda (sexp)  
    (make-body (map parse-exp sexp))))

;; parse "expression" in "method body" : make AST.
(define parse-exp
  (lambda (sexp)
    (cond ((or (number? sexp)   ;; number or
	       (string? sexp)   ;; string or 
	       (eqv? sexp #t)   ;; true of boolean or
	       (eqv? sexp #f))  ;; false of boolean
	   (make-lit-exp sexp)) ;; is "literal"

	  ((and (list? sexp) (eqv? (car sexp) 'quote)) 
	   (make-lit-exp (cadr sexp)))	               
       
	  ((symbol? sexp) (make-var-get-exp sexp))
	  
	  ((list? sexp)

	   ;; prim is the type whose key matchs 1st of sexp
	   ;;   -> the result is lambda adstract(= procedure)
	   ;;;; "assq" is the procedure to get object from "association list":
	   ;;;;   it uses "eq?" to test the equality. 
	   ;;;; "eq?" is the test judge that two pointers are the same.
	   (let ((prim (assq (car sexp) *primitives*))) 
	     (if prim
		 (make-prim-app-exp (cadr prim) (parse-rands (cdr sexp))) 

		 (case (car sexp) 
		   
		   ((if) 
		    (make-if-exp (parse-exp (cadr sexp))     ;; condition 
				 (parse-exp (caddr sexp))    ;; then-clause
				 (parse-exp (cadddr sexp)))) ;; else-clause

		   ((let) 
		    (make-let-exp (parse-params (map car (cadr sexp)))
				  (parse-rands (map cadr (cadr sexp)))
				  (parse-body (cddr sexp))))
		   
		   ((set!) 
		    (make-var-set-exp (cadr sexp) (parse-exp (caddr sexp))))

		   ;; the following part : oo language

		   ((fget) 
		    (make-field-get-exp (parse-exp (cadr sexp)) (caddr sexp)))

		   ((fset!)
		    (make-field-set-exp (parse-exp (cadr sexp))
						(caddr sexp)
						(parse-exp (cadddr sexp))))

		   ((send) 
		    (make-method-call-exp (parse-exp (cadr sexp))
					  (caddr sexp)
					  (parse-rands (cdddr sexp))))

		   ((super)
		    (make-super-call-exp (cadr sexp) (parse-rands (cddr sexp))))

		   ((new) 
		    (make-new-obj-exp (cadr sexp) (parse-rands (cddr sexp))))

		   ((instanceof?) 
		    (make-instanceof-exp (cadr sexp) (parse-exp (caddr sexp))))

		   ;; the following part : aspects
		   ((proceed) 
		    (make-proceed-exp (parse-rands (cdr sexp))))

		   ;; default case : error
		   (else
		    (error 'parse-exp "not an expression -- ~s" sexp))))))

	  (else (error 'parse-exp "not an expression -- ~s" sexp)))))

(define parse-exps  
  (lambda (exps) (map parse-exp exps)))

;; parse "rand" - what is "rand"?
(define parse-rands
  parse-exps)

(define parse-pcd
  (lambda (sexp)
    (if (not (memq *semantics* '(pa paex padflow trav travex)));if you ever want to
						       ;have traversals with
					               ;moby-ajd then you
						       ;deserve to lose for
						       ;trying to do such hair!
	(error "Can't parse pcd in this semantics."))
    
    ;; case of 'joinpoint'
    (case (car sexp)
      ((call) 
       (parse-method-signature-pcd make-call-pcd sexp))

      ((mexecution) 
       (parse-method-signature-pcd make-mexecution-pcd sexp))

      ((aexecution) 
       (make-aexecution-pcd))

      ((new) ;; "call/execution of constructor"
       (if (eq? *semantics* 'moby-ajd)
	   (make-new-pcd (cadr sexp) (caddr sexp))
	   (make-new-pcd (cadr sexp) #f)))

      ((get) 
       (make-get-pcd (cadr sexp)))
      ((set) 
       (make-set-pcd (cadr sexp)))

      ((cwithin) ;; "within(Class)"
       (make-cwithin-pcd (cadr sexp)))
      ((mwithin) ;; "within(Method)"
       (make-mwithin-pcd (cadr sexp)))


      ((this) 
       (make-this-pcd   (parse-param (cadr sexp))))
      ((target) 
       (make-target-pcd (parse-param (cadr sexp))))
      ((args) 
       (make-args-pcd   (parse-params (cdr sexp))))

      ((return) 
       (make-return-pcd   (parse-param (cadr sexp)))) ;; kazu

      ((cflow) 
       (make-cflow-pcd      (parse-pcd (cadr sexp))))
      ((cflowtop) 
       (make-cflowtop-pcd   (parse-pcd (cadr sexp))))
      ((cflowbelow) 
       (make-cflowbelow-pcd (parse-pcd (cadr sexp))))

      ;; [ver1] dflow (vs,vd,pcd) 
      ;; -> [ver2] dflow (v-dest pc-dest, v-src pc-src)
      ;; -> [ver3] dflow (v-dest pc-dest, v-src pc-src, &bypassing &v-bypass &pc-bypass)
      ((dflow)
       (if (null? (cdddddr sexp))  ;; "bypassing" clause is null?
	   (make-dflow-pcd (cadr sexp)     (parse-pcd (caddr sexp))          ; 2nd, 3rd (1st is 'dflow')
			   (cadddr sexp)   (parse-pcd (caddddr sexp))        ; 4th, 5th
			   '*void-bypass-value* '*void-bypass-pcd*)
	   ;; "bypassing"
	   (make-dflow-pcd (cadr sexp)      (parse-pcd (caddr sexp))         ; 2nd, 3rd
			   (cadddr sexp)    (parse-pcd (caddddr sexp))       ; 4th, 5th
			   (caddddddr sexp) (parse-pcd (cadddddddr sexp))))) ; 7th, 8th (6th is 'bypassing')

      ((morerecent) 
       (make-morerecent-pcd (parse-pcd (cadr sexp)) (parse-pcd (caddr sexp))))

      ((traversal) 
       (make-traversal-pcd  (cadr sexp)))
      ((crossing) 
       (make-crossing-pcd   (cadr sexp)))

      ((and) 
       (make-and-pcd (parse-pcds (cdr sexp))))
      ((or) 
       (make-or-pcd  (parse-pcds (cdr sexp))))
      ((not) 
       (make-not-pcd (parse-pcd  (cadr sexp))))

      (else
       (make-named-pcd 
	(car sexp) (parse-params (cdr sexp)))))))

;; parse "method signature of pointcut-designator"
(define parse-method-signature-pcd
  (lambda (maker sexp)
    (cond ((and (eq? *semantics* 'moby-ajd)
		(= (length sexp) 5))
	   (maker (cadr sexp) (caddr sexp) (cadddr sexp) (caddddr sexp)))
	  ((= (length sexp) 2)
	   (maker #f          #f           (cadr sexp)   #f))
	  (else
	   (error 'parse-pcd "Can't parse ~s." sexp)))))


(define parse-pcds 
  (lambda (sexps) (map parse-pcd sexps)))

;;;;;;;;;;;;;;;;;;;;;;
;; walk(= traverse) : for making "... declaration"
;;;;;;;;;;;;;;;;;;;;;;
(define walk-program
  (lambda (pgm proc)
    (let ((new-pgm (proc pgm)))
      (make-program (walk-decls (program-decls new-pgm) proc)))))

(define walk-decls
  (lambda (decls proc)
    (remove* #f (map (lambda (decl) (walk-decl decl proc)) decls))))

(define walk-exps
  (lambda (exps proc)
    (remove #f (map (lambda (exp) (walk-exp exp proc)) exps))))

(define walk-rands
  (lambda (exps proc)
    (map (lambda (exp) (walk-exp exp proc)) exps)))

(define walk-decl
  (lambda (decl proc)
    (let ((decl (proc decl)))
      (cond ((class-decl? decl)
	     (make-class-decl  (class-decl-cname decl)
			       (class-decl-sname decl)
			       (walk-decls (class-decl-decls decl) proc)))
	    ((field-decl? decl)
	     (make-field-decl  (member-decl-cname decl)
			       (field-decl-type   decl)
			       (field-decl-fname   decl)))	      
	    ((method-decl? decl)
	     (make-method-decl (member-decl-cname decl)
			       (method-decl-rtype  decl)
			       (method-decl-mname   decl)
			       (method-decl-params decl)
			       (walk-body (method-decl-body decl) proc)))
	    ((advice-decl? decl)
	     (make-advice-decl (member-decl-cname decl)
			       (advice-decl-qualifier  decl)
			       (advice-decl-params decl)
			       (walk-pcd  (advice-decl-pcd  decl) proc)
			       (walk-body (advice-decl-body decl) proc)))
	    ((eq? decl #f) decl)	; proc may return #f to erase the decl
	    (else
	     (error 'walk-decl "What is ~s?" decl))))))

(define walk-body
  (lambda (bdy proc)
    (let ((new-bdy (proc bdy)))
      (cond ((body? new-bdy) (make-body (walk-exps (body-exps new-bdy) proc)))
	    (else            new-bdy)))))

(define walk-exp
  (lambda (exp proc)
    (let ((new-exp (proc exp)))
      (cond ((lit-exp? new-exp) new-exp)
          
	    ((if-exp? new-exp)
	     (make-if-exp (walk-exp (if-exp-test-exp new-exp) proc)
	                     (walk-exp (if-exp-then-exp new-exp) proc)
			     (walk-exp (if-exp-else-exp new-exp) proc)))
     
	    ((prim-app-exp? new-exp)
	     (make-prim-app-exp (prim-app-exp-prim new-exp)
	                           (walk-rands (prim-app-exp-rands new-exp) proc)))
     
	    ((and-exp? new-exp)
	     (make-and-exp (walk-rands (and-exp-rands new-exp) proc)))
	    
	    ((or-exp? new-exp)
	     (make-or-exp  (walk-rands (or-exp-rands new-exp) proc)))
     
	    ((let-exp? new-exp)
	     (make-let-exp (let-exp-params new-exp)
	                      (walk-rands (let-exp-rands new-exp) proc)
			      (walk-body (let-exp-body new-exp) proc)))
	    
	    ((var-get-exp? new-exp)
	     (make-var-get-exp (var-get-exp-vname new-exp)))
     
	    ((var-set-exp? new-exp)
	     (make-var-set-exp (var-set-exp-vname new-exp)
			       (walk-exp (var-set-exp-rhs-exp new-exp) proc)))

	    ((method-call-exp? new-exp)
	     (make-method-call-exp 
	      (walk-exp  (method-call-exp-obj-exp new-exp) proc)
	      (method-call-exp-mname new-exp)
	      (walk-rands (method-call-exp-rands new-exp) proc)))
     
	    ((super-call-exp? new-exp)
	     (make-super-call-exp
	      (super-call-exp-mname new-exp)
	      (walk-rands (super-call-exp-rands new-exp) proc)))
     
	    ((new-obj-exp? new-exp)
	     (make-new-obj-exp 
	      (new-obj-exp-cname new-exp)
	      (walk-rands (new-obj-exp-rands new-exp) proc)))


	    ((field-get-exp? new-exp)
	     (make-field-get-exp 
	      (walk-exp (field-get-exp-obj-exp new-exp) proc)
	      (field-get-exp-fname new-exp)))

	    ((field-set-exp? new-exp)
	     (make-field-set-exp
	      (walk-exp (field-set-exp-obj-exp new-exp) proc)
	      (field-set-exp-fname new-exp)
	      (walk-exp (field-set-exp-rhs-exp new-exp) proc)))

	    ((instanceof-exp? new-exp)
	     (make-instanceof-exp
	      (instanceof-exp-cname new-exp)
	      (walk-exp (instanceof-exp-obj-exp new-exp) proc)))

	    ;; aspects
	    ((proceed-exp? new-exp)
	     (make-proceed-exp
	       (walk-rands (proceed-exp-rands new-exp) proc)))

	    (else (error 'walk-exp "not an expression -- ~s" exp))))))

(define walk-pcd
  (lambda (pcd proc)
    (let ((new-pcd (proc pcd)))
      (cond ((or (call-pcd? new-pcd)
		 (mexecution-pcd? new-pcd)

		 (new-pcd? new-pcd)

		 (get-pcd? new-pcd)
		 (set-pcd? new-pcd)
		 
		 (cwithin-pcd? new-pcd)
		 (mwithin-pcd? new-pcd)
		 
		 (this-pcd? new-pcd)
		 (target-pcd? new-pcd)
		 (args-pcd? new-pcd)

		 (return-pcd? new-pcd) ;; kazu

		 (named-pcd? new-pcd)
	       
		 (traversal-pcd? new-pcd)
		 (crossing-pcd? new-pcd))
	     new-pcd)

	  ((cflow-pcd? new-pcd)
	   (make-cflow-pcd      (walk-pcd (cflow-pcd-pcd new-pcd) proc)))
	  ((cflowbelow-pcd? new-pcd)
	   (make-cflowbelow-pcd (walk-pcd (cflowbelow-pcd-pcd new-pcd) proc)))
	  ((cflowtop-pcd? new-pcd)
	   (make-cflowtop-pcd   (walk-pcd (cflowtop-pcd-pcd new-pcd) proc)))

	  ;; for dflow 
	  ; ((dflow-pcd? new-pcd)
	  ;  (make-dflow-pcd 
	  ;   ( ??? ) ))

	  ((morerecent-pcd? new-pcd)
	   (make-morerecent-pcd (walk-pcd (morerecent-pcd-pcd new-pcd) proc)))

	  ((and-pcd? new-pcd)
	   (make-and-pcd (map (lambda (sub) (walk-pcd sub proc)) (and-pcd-pcds new-pcd))))
	  ((or-pcd? new-pcd)
	   (make-or-pcd  (map (lambda (sub) (walk-pcd sub proc)) (or-pcd-pcds  new-pcd))))
	  ((not-pcd? new-pcd)
	   (make-not-pcd (walk-pcd (not-pcd-pcd new-pcd) proc)))))))


(define remove-decls-if
  (lambda (decls pred)
    (cond ((null? decls) decls)
	  ((pred (car decls)) (remove-decls-if (cdr decls) pred))
	  ((class-decl? (car decls))
	   (cons (make-class-decl (class-decl-cname (car decls))
				  (class-decl-sname (car decls))
				  (remove-decls-if (class-decl-decls (car decls))
						   pred))
		 (remove-decls-if (cdr decls) pred)))
	  (else
	   (cons (car decls)
		 (remove-decls-if (cdr decls) pred))))))



;;;
;;; primitives
;;;
(define *primitives*
  `((+       ,(lambda args (apply + args)))
    (-       ,(lambda args (apply - args)))
    (*       ,(lambda args (apply * args)))
    (/       ,(lambda args (apply / args)))
    (=       ,(lambda args (apply = args)))
    (<       ,(lambda args (apply < args)))
    (>       ,(lambda args (apply > args)))
    (zero?   ,(lambda args (apply zero? args)))

    (not     ,(lambda args (apply not args)))

    (write   ,(lambda args (apply write args)))
    (newline ,(lambda args (apply newline args)))

    (car     ,(lambda args (apply car  args)))
    (cdr     ,(lambda args (apply cdr  args)))
    (cons    ,(lambda args (apply cons args)))
    (list    ,(lambda args args))
    (length  ,(lambda args (apply length args)))
    (null?   ,(lambda args (apply null? args)))
    ;; for string
    (string-append  ,(lambda args (apply string-append args)))
    (string->number ,(lambda args (apply string->number args)))
    (number->string ,(lambda args (apply number->string args)))
    ))
